import abc
import contextlib
import inspect
import typing
from contextlib import contextmanager
from operator import attrgetter

import typing_extensions
from typing_extensions import override

from that_depends.entities.resource_context import ResourceContext


T_co = typing.TypeVar("T_co", covariant=True)
R = typing.TypeVar("R")
P = typing.ParamSpec("P")
ResourceCreatorType: typing.TypeAlias = typing.Callable[
    P,
    typing.Iterator[T_co] | typing.AsyncIterator[T_co] | typing.ContextManager[T_co] | typing.AsyncContextManager[T_co],
]


class AbstractProvider(typing.Generic[T_co], abc.ABC):
    """Base class for all providers."""

    def __init__(self, **kwargs: typing.Any) -> None:  # noqa: ANN401
        """Create a new provider."""
        super().__init__(**kwargs)
        self._override: typing.Any = None

    def __deepcopy__(self, *_: object, **__: object) -> typing_extensions.Self:
        """Hack for Litestar to prevent cloning object.

        More info here https://github.com/modern-python/that-depends/issues/119.
        """
        return self

    def __getattr__(self, attr_name: str) -> typing.Any:  # noqa: ANN401
        """Get an attribute from the resolve object.

        Args:
            attr_name: name of attribute to get.

        Returns:
            An `AttrGetter` provider that will get the attribute after resolving the current provider.

        """
        if attr_name.startswith("_"):
            msg = f"'{type(self)}' object has no attribute '{attr_name}'"
            raise AttributeError(msg)
        return AttrGetter(provider=self, attr_name=attr_name)

    @abc.abstractmethod
    async def async_resolve(self) -> T_co:
        """Resolve dependency asynchronously."""

    @abc.abstractmethod
    def sync_resolve(self) -> T_co:
        """Resolve dependency synchronously."""

    async def __call__(self) -> T_co:
        """Resolve dependency asynchronously."""
        return await self.async_resolve()

    def override(self, mock: object) -> None:
        """Override the provider with a mock object.

        Args:
            mock: object to resolve while the provider is overridden.

        Returns:
            None

        """
        self._override = mock

    @contextmanager
    def override_context(self, mock: object) -> typing.Iterator[None]:
        """Override the provider with a mock object temporarily.

        Args:
            mock: object to resolve while the provider is overridden.

        Returns:
            None

        """
        self.override(mock)
        try:
            yield
        finally:
            self.reset_override()

    def reset_override(self) -> None:
        """Reset the provider to its original state.

        Use this is you have previously called `override` or `override_context`
        to reset the provider to its original state.
        """
        self._override = None

    @property
    def cast(self) -> T_co:
        """Returns self, but cast to the type of the provided value.

        This helps to pass providers as input to other providers while avoiding type checking errors:

        Example:
            class A: ...

            def create_b(a: A) -> B: ...

            class Container(BaseContainer):
                a_factory = Factory(A)
                b_factory1 = Factory(create_b, a_factory)  # works, but mypy (or pyright, etc.) will complain
                b_factory2 = Factory(create_b, a_factory.cast)  # works and passes type checking

        """
        return typing.cast(T_co, self)


class AbstractResource(AbstractProvider[T_co], abc.ABC):
    """Base class for Resource providers."""

    def __init__(
        self,
        creator: ResourceCreatorType[P, T_co],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """Create a new resource.

        Args:
            creator: sync or async iterator or context manager that yields resource.
            *args: positional arguments to pass to the creator.
            **kwargs: keyword arguments to pass to the creator.


        """
        super().__init__()
        self._creator: typing.Any

        if inspect.isasyncgenfunction(creator):
            self.is_async = True
            self._creator = contextlib.asynccontextmanager(creator)
        elif inspect.isgeneratorfunction(creator):
            self.is_async = False
            self._creator = contextlib.contextmanager(creator)
        elif isinstance(creator, type) and issubclass(creator, typing.AsyncContextManager):
            self.is_async = True
            self._creator = creator
        elif isinstance(creator, type) and issubclass(creator, typing.ContextManager):
            self.is_async = False
            self._creator = creator
        else:
            msg = "Unsupported resource type"
            raise TypeError(msg)
        self._args: P.args = args
        self._kwargs: P.kwargs = kwargs

    @abc.abstractmethod
    def _fetch_context(self) -> ResourceContext[T_co]: ...

    @override
    async def async_resolve(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        context = self._fetch_context()

        if context.instance is not None:
            return context.instance

        # lock to prevent race condition while resolving
        async with context.asyncio_lock:
            if context.instance is not None:
                return context.instance

            cm: typing.ContextManager[T_co] | typing.AsyncContextManager[T_co] = self._creator(
                *[await x.async_resolve() if isinstance(x, AbstractProvider) else x for x in self._args],
                **{
                    k: await v.async_resolve() if isinstance(v, AbstractProvider) else v
                    for k, v in self._kwargs.items()
                },
            )

            if isinstance(cm, typing.AsyncContextManager):
                context.context_stack = contextlib.AsyncExitStack()
                context.instance = await context.context_stack.enter_async_context(cm)

            elif isinstance(cm, typing.ContextManager):
                context.context_stack = contextlib.ExitStack()
                context.instance = context.context_stack.enter_context(cm)

            else:  # pragma: no cover
                typing.assert_never(cm)

        return context.instance

    @override
    def sync_resolve(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        context = self._fetch_context()
        if context.instance is not None:
            return context.instance

        if self.is_async:
            msg = "AsyncResource cannot be resolved synchronously"
            raise RuntimeError(msg)

        # lock to prevent race condition while resolving
        with context.threading_lock:
            if context.instance is not None:
                return context.instance

            cm = self._creator(
                *[x.sync_resolve() if isinstance(x, AbstractProvider) else x for x in self._args],
                **{k: v.sync_resolve() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()},
            )
            context.context_stack = contextlib.ExitStack()
            context.instance = context.context_stack.enter_context(cm)

            return context.instance


def _get_value_from_object_by_dotted_path(obj: typing.Any, path: str) -> typing.Any:  # noqa: ANN401
    attribute_getter = attrgetter(path)
    return attribute_getter(obj)


class AttrGetter(
    AbstractProvider[T_co],
):
    """Provides an attribute after resolving the wrapped provider."""

    __slots__ = "_attrs", "_provider"

    def __init__(self, provider: AbstractProvider[T_co], attr_name: str) -> None:
        """Create a new AttrGetter instance.

        Args:
            provider: provider to wrap.
            attr_name: attribute name to resolve when the provider is resolved.

        """
        super().__init__()
        self._provider = provider
        self._attrs = [attr_name]

    @override
    def __getattr__(self, attr: str) -> "AttrGetter[T_co]":
        if attr.startswith("_"):
            msg = f"'{type(self)}' object has no attribute '{attr}'"
            raise AttributeError(msg)
        self._attrs.append(attr)
        return self

    @override
    async def async_resolve(self) -> typing.Any:
        resolved_provider_object = await self._provider.async_resolve()
        attribute_path = ".".join(self._attrs)
        return _get_value_from_object_by_dotted_path(resolved_provider_object, attribute_path)

    @override
    def sync_resolve(self) -> typing.Any:
        resolved_provider_object = self._provider.sync_resolve()
        attribute_path = ".".join(self._attrs)
        return _get_value_from_object_by_dotted_path(resolved_provider_object, attribute_path)



import functools
import inspect
import re
import typing
import warnings

from that_depends import BaseContainer
from that_depends.meta import BaseContainerMeta
from that_depends.providers import AbstractProvider
from that_depends.providers.context_resources import ContextScope, ContextScopes, container_context


P = typing.ParamSpec("P")
T = typing.TypeVar("T")


@typing.overload
def inject(func: typing.Callable[P, T]) -> typing.Callable[P, T]: ...


@typing.overload
def inject(
    *,
    scope: ContextScope | None = ContextScopes.INJECT,
) -> typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]: ...


def inject(  # noqa: C901
    func: typing.Callable[P, T] | None = None, scope: ContextScope | None = ContextScopes.INJECT
) -> typing.Callable[P, T] | typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]:
    """Inject dependencies into a function."""
    if scope == ContextScopes.ANY:
        msg = f"{scope} is not allowed in inject decorator."
        raise ValueError(msg)

    def _inject(
        func: typing.Callable[P, T],
    ) -> typing.Callable[P, T]:
        if inspect.iscoroutinefunction(func):
            return typing.cast(typing.Callable[P, T], _inject_to_async(func))

        return _inject_to_sync(func)

    def _inject_to_async(
        func: typing.Callable[P, typing.Coroutine[typing.Any, typing.Any, T]],
    ) -> typing.Callable[P, typing.Coroutine[typing.Any, typing.Any, T]]:
        @functools.wraps(func)
        async def inner(*args: P.args, **kwargs: P.kwargs) -> T:
            if scope:
                async with container_context(scope=scope):
                    return await _resolve_async(func, *args, **kwargs)
            return await _resolve_async(func, *args, **kwargs)

        return inner

    def _inject_to_sync(
        func: typing.Callable[P, T],
    ) -> typing.Callable[P, T]:
        @functools.wraps(func)
        def inner(*args: P.args, **kwargs: P.kwargs) -> T:
            if scope:
                with container_context(scope=scope):
                    return _resolve_sync(func, *args, **kwargs)
            return _resolve_sync(func, *args, **kwargs)

        return inner

    if func:
        return _inject(func)

    return _inject


def _resolve_sync(func: typing.Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
    injected = False
    signature: typing.Final = inspect.signature(func)
    for i, (field_name, field_value) in enumerate(signature.parameters.items()):
        if i < len(args):
            continue
        if not isinstance(field_value.default, AbstractProvider) and not isinstance(
            field_value.default, StringProviderDefinition
        ):
            continue
        if field_name in kwargs:
            if isinstance(field_value.default, AbstractProvider | StringProviderDefinition):
                injected = True
            continue
        if isinstance(field_value.default, StringProviderDefinition):
            kwargs[field_name] = field_value.default.provider.sync_resolve()
        else:
            kwargs[field_name] = field_value.default.sync_resolve()
        injected = True

    if not injected:
        warnings.warn("Expected injection, but nothing found. Remove @inject decorator.", RuntimeWarning, stacklevel=1)

    return func(*args, **kwargs)


async def _resolve_async(
    func: typing.Callable[P, typing.Coroutine[typing.Any, typing.Any, T]], *args: P.args, **kwargs: P.kwargs
) -> T:
    injected = False
    signature = inspect.signature(func)
    for i, (field_name, field_value) in enumerate(signature.parameters.items()):
        if i < len(args):
            if isinstance(field_value.default, AbstractProvider | StringProviderDefinition):
                injected = True
            continue

        if not isinstance(field_value.default, AbstractProvider) and not isinstance(
            field_value.default, StringProviderDefinition
        ):
            continue

        if field_name in kwargs:
            if isinstance(field_value.default, AbstractProvider | StringProviderDefinition):
                injected = True
            continue
        if isinstance(field_value.default, StringProviderDefinition):
            kwargs[field_name] = await field_value.default.provider.async_resolve()
        else:  # AbstractProvider
            kwargs[field_name] = await field_value.default.async_resolve()
        injected = True
    if not injected:
        warnings.warn("Expected injection, but nothing found. Remove @inject decorator.", RuntimeWarning, stacklevel=1)
    return await func(*args, **kwargs)


class StringProviderDefinition:
    """Provider definition from a string."""

    def __init__(self, definition: str) -> None:
        """Initialize the provider definition.

        Args:
            definition: provider definition in the format container.provider[.attr1.attr2...].

        """
        self._definition = definition
        self._container_name, self._provider_name, self._attrs = self._validate_and_extract_provider_definition()

    def _validate_and_extract_provider_definition(self) -> tuple[str, str, list[str]]:
        pattern = r"^([^.]+)\.([^.]+)(?:\.(.+))?$"
        match = re.match(pattern, self._definition)
        if match:
            container_name = match.group(1)
            provider_name = match.group(2)
            attrs = match.group(3).split(".") if match.group(3) else []
            return container_name, provider_name, attrs
        msg = f"Invalid provider definition: {self._definition}"
        raise ValueError(msg)

    def _get_container_by_name(self) -> type["BaseContainer"]:
        containers_in_scope = BaseContainerMeta.get_instances()
        try:
            return containers_in_scope[self._container_name]
        except KeyError as e:
            msg = f"Container {self._container_name} not found in scope!"
            raise ValueError(msg) from e

    def _get_provider_by_name(self) -> AbstractProvider[typing.Any]:
        container = self._get_container_by_name()
        try:
            provider = container.get_providers()[self._provider_name]
        except KeyError as e:
            msg = f"Provider {self._provider_name} not found in container {self._container_name}"
            raise ValueError(msg) from e
        for attr in self._attrs:
            provider = getattr(provider, attr)
        return provider

    @property
    def provider(self) -> AbstractProvider[typing.Any]:
        """Get the provider instance."""
        return self._get_provider_by_name()


class ClassGetItemMeta(type):
    """Metaclass to support Provide[provider] syntax."""

    def __getitem__(cls, provider: AbstractProvider[T] | str) -> T | typing.Any:  # noqa: ANN401
        if isinstance(provider, str):
            return StringProviderDefinition(provider)  # will be resolved later
        return typing.cast(T, provider)


class Provide(metaclass=ClassGetItemMeta):
    """Marker to dependency injection."""



import typing

from that_depends.entities.resource_context import ResourceContext
from that_depends.providers.base import AbstractResource, ResourceCreatorType


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class Resource(AbstractResource[T_co]):
    """Provides a resource that is resolved once and cached for future usage.

    Unlike a singleton, this provider includes finalization logic and can be
    used with a generator or async generator to manage resource lifecycle.
    It also supports usage with `typing.ContextManager` or `typing.AsyncContextManager`.
    Threading and asyncio concurrency are supported, ensuring only one instance
    is created regardless of concurrent resolves.

    Example:
        ```python
        async def create_async_resource():
            try:
                yield "async resource"
            finally:
                # Finalize resource
                pass

        class MyContainer:
            async_resource = Resource(create_async_resource)

        async def main():
            async_resource_instance = await MyContainer.async_resource.async_resolve()
            await MyContainer.async_resource.tear_down()
        ```

    """

    __slots__ = (
        "_args",
        "_context",
        "_creator",
        "_creator",
        "_is_async",
        "_kwargs",
        "_override",
        "is_async",
    )

    def __init__(
        self,
        creator: ResourceCreatorType[P, T_co],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """Initialize the Resource provider with a callable for resource creation.

        The callable can be a generator or async generator that yields the resource
        (with optional teardown logic), or a context manager. Only one instance will be
        created and cached until explicitly torn down.

        Args:
            creator: The callable, generator, or context manager that creates the resource.
            *args: Positional arguments passed to the creator.
            **kwargs: Keyword arguments passed to the creator.

        Example:
            ```python
            def custom_creator(name: str):
                try:
                    yield f"Resource created for {name}"
                finally:
                    pass  # Teardown

            resource_provider = Resource(custom_creator, "example")
            instance = resource_provider.sync_resolve()
            resource_provider.tear_down()
            ```

        """
        super().__init__(creator, *args, **kwargs)
        self._context: typing.Final[ResourceContext[T_co]] = ResourceContext(is_async=self.is_async)

    def _fetch_context(self) -> ResourceContext[T_co]:
        return self._context

    async def tear_down(self) -> None:
        """Tear down the resource if it has been created.

        If the resource was never resolved, or was already torn down,
        calling this method has no effect.

        Example:
            ```python
            # Assuming my_provider was previously resolved
            await my_provider.tear_down()
            ```

        """
        await self._fetch_context().tear_down()



import abc
import typing
from collections.abc import MutableMapping
from threading import Lock

from typing_extensions import override


if typing.TYPE_CHECKING:
    from that_depends.container import BaseContainer


class DefaultScopeNotDefinedError(Exception):
    """Exception raised when default_scope is not defined."""


class _ContainerMetaDict(dict[str, typing.Any]):
    """Implements custom logic for the container metaclass."""

    @override
    def __setitem__(self, key: str, value: typing.Any) -> None:
        from that_depends.providers.context_resources import ContextResource, ContextScopes

        if isinstance(value, ContextResource) and value.get_scope() == ContextScopes.ANY:
            try:
                default_scope = self.__getitem__("default_scope")
                super().__setitem__(key, value.with_config(default_scope))
            except KeyError as e:
                msg = "Explicitly define default_scope before defining ContextResource providers."
                raise DefaultScopeNotDefinedError(msg) from e
        else:
            super().__setitem__(key, value)


class BaseContainerMeta(abc.ABCMeta):
    """Metaclass for BaseContainer."""

    _instances: typing.ClassVar[dict[str, type["BaseContainer"]]] = {}

    _MUTABLE_ATTRS = (
        "__abstractmethods__",
        "__parameters__",
        "_abc_impl",
        "providers",
        "containers",
    )

    _lock: Lock = Lock()

    def name(cls) -> str:
        """Get the name of the container class."""
        if alias := getattr(cls, "alias", None):
            return typing.cast(str, alias)
        return cls.__name__

    @classmethod
    @override
    def __prepare__(cls, name: str, bases: tuple[type, ...], /, **kwds: typing.Any) -> MutableMapping[str, object]:
        return _ContainerMetaDict()

    @override
    def __new__(cls, name: str, bases: tuple[type, ...], namespace: dict[str, typing.Any]) -> type:
        new_cls = super().__new__(cls, name, bases, namespace)
        cls_name = new_cls.name()
        with cls._lock:
            if name != "BaseContainer":
                cls._instances[cls_name] = typing.cast(type["BaseContainer"], new_cls)
        return new_cls

    @classmethod
    def get_instances(cls) -> dict[str, type["BaseContainer"]]:
        """Get all instances that inherit from BaseContainer."""
        return cls._instances

    @override
    def __setattr__(cls, key: str, value: typing.Any) -> None:
        if key in cls._MUTABLE_ATTRS:  # Allow modification of mutable attributes
            super().__setattr__(key, value)
        else:
            msg = f"Cannot add new attribute '{key}' to class '{cls.__name__}'"
            raise AttributeError(msg)



"""Selection based providers."""

import typing

from typing_extensions import override

from that_depends.providers.base import AbstractProvider


T_co = typing.TypeVar("T_co", covariant=True)


class Selector(AbstractProvider[T_co]):
    """Chooses a provider based on a key returned by a selector function.

    This class allows you to dynamically select and resolve one of several
    named providers at runtime. The provider key is determined by a
    user-supplied selector function.

    Examples:
        ```python
        def environment_selector():
            return "local"

        selector_instance = Selector(
            environment_selector,
            local=LocalStorageProvider(),
            remote=RemoteStorageProvider(),
        )

        # Synchronously resolve the selected provider
        service = selector_instance.sync_resolve()
        ```

    """

    __slots__ = "_override", "_providers", "_selector"

    def __init__(self, selector: typing.Callable[[], str], **providers: AbstractProvider[T_co]) -> None:
        """Initialize a new Selector instance.

        Args:
            selector (Callable[[], str]): A function that returns the key
                of the provider to use.
            **providers (AbstractProvider[T_co]): The named providers from
                which one will be selected based on the `selector`.

        Examples:
            ```python
            def my_selector():
                return "remote"

            my_selector_instance = Selector(
                my_selector,
                local=LocalStorageProvider(),
                remote=RemoteStorageProvider(),
            )

            # The "remote" provider will be selected
            selected_service = my_selector_instance.sync_resolve()
            ```

        """
        super().__init__()
        self._selector: typing.Final = selector
        self._providers: typing.Final = providers

    @override
    async def async_resolve(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        selected_key: typing.Final = self._selector()
        if selected_key not in self._providers:
            msg = f"No provider matches {selected_key}"
            raise RuntimeError(msg)
        return await self._providers[selected_key].async_resolve()

    @override
    def sync_resolve(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        selected_key: typing.Final = self._selector()
        if selected_key not in self._providers:
            msg = f"No provider matches {selected_key}"
            raise RuntimeError(msg)
        return self._providers[selected_key].sync_resolve()
